/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.common.json.models;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.annotations.SerializedName;

import com.redhat.thermostat.common.yaml.JsonToYaml;
import com.redhat.thermostat.lang.schema.annotations.Schema;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Partial Java representation of a JSON Schema <br>
 * (see json-schema.org)
 */
public class SchemaDef implements SchemaType {

    private static final Gson gson;
    private static final int DEFAULT_MAXDEPTH = 1;

    static {
        final boolean ENABLE_PRETTY_PRINT = false;
        if (ENABLE_PRETTY_PRINT) {
            GsonBuilder gbuilder = new GsonBuilder();
            gbuilder.setPrettyPrinting();
            gson = gbuilder.create();
        } else {
            gson = new Gson();
        }
    }

    @SerializedName("$schema")
    private String schema;
    @SerializedName("$id")
    private String id;
    private String title;
    private String description;
    private Map<String, SchemaType> properties;
    private List<String> required;

    /** standard JSON schema properties that are not yet implemented:
     private int multipleOf;
     private long maximum;
     private long exclusiveMaximum;
     private long minimum;
     private long exclusiveMinimum;
     private int maxLength;
     private int minLength;
     private String pattern;
     private int maxItems;
     private int minItems;
     private boolean uniqueItems;
     private int maxProperties;
     private int minProperties;
     (and more, see json-schema.org)
     ***/

    // for GSon
    public SchemaDef() {}

    public SchemaDef(Object target) {
        this(target.getClass(), DEFAULT_MAXDEPTH);
    }

    public SchemaDef(Object target, int maxDepth) {
        this(target.getClass(), maxDepth);
    }

    public SchemaDef(Class target) {
        this(target, true, DEFAULT_MAXDEPTH);
    }

    public SchemaDef(Class target, int maxDepth) {
        this(target, true, maxDepth);
    }

    private SchemaDef(Class target, boolean rootLevel, int maxDepth) {
        if (rootLevel) {
            schema = "http://json-schema.org/draft-04/schema#";
        }
        Field[] fields = target.getDeclaredFields();
        if (fields.length > 0) {
            properties = new HashMap<>(fields.length);
            required = new ArrayList<>();
            for (Field field : fields) {
                final boolean isTransient =  Modifier.isTransient(field.getModifiers());
                if (!isTransient) {
                    String fname = field.getAnnotation(SerializedName.class) != null ? field.getAnnotation(SerializedName.class).value() : field.getName();
                    Schema schema = field.getAnnotation(Schema.class);
                    String descr = null;
                    if (schema != null) {
                        if (!schema.name().isEmpty()) {
                            fname = schema.name();
                        }
                        descr = schema.description();
                        if (schema.required()) {
                            required.add(fname);
                        }
                    }
                    if (field.getType().isPrimitive()) {
                        String tt = field.getType().getSimpleName();
                        properties.put(fname, new SimpleType(tt, descr));
                    } else if (maxDepth > 1){
                        properties.put(fname, new SchemaDef(field.getType(), false, maxDepth - 1));
                    } else {
                        properties.put(fname, new SimpleType(field.getType().getSimpleName(), descr));
                    }
                }
            }
            if (required.isEmpty()) {
                required = null;
            }
        }
    }

    public Map<String, SchemaType> getProperties() {
        return this.properties;
    }

    public List<String> getRequired() {
        return this.required;
    }

    public JsonObject toJsonObject() {
        JsonElement el = gson.toJsonTree(this);
        return el.isJsonObject() ? el.getAsJsonObject() : null;
    }

    public String toString() {
        return gson.toJson(this);
    }

    @Override
    public TypeClass getTypeClass() {
        return TypeClass.OBJECT;
    }

    public static class SimpleType implements SchemaType {
        transient TypeClass type;
        @SerializedName("type")
        String typeStr;
        String description;
        SimpleType(String tname, String descr) {
            if ("string".equals(tname)) {
                type = TypeClass.STRING;
            } else if ("int".equals(tname) || "float".equals(tname) || "double".equals(tname) || "long".equals(tname) || "number".equals(tname)) {
                type = TypeClass.NUMBER;
            } else if ("Integer".equals(tname) || "Float".equals(tname) || "Double".equals(tname) || "Long".equals(tname)) {
                type = TypeClass.NUMBER;
            } else if ("boolean".equals(tname) || "Boolean".equals(tname)) {
                type = TypeClass.BOOLEAN;
            } else {
                type = TypeClass.OBJECT;
            }
            this.typeStr = type == TypeClass.OBJECT ? tname : type.toString();
            this.description = descr;
        }
        @Override
        public TypeClass getTypeClass() {
            return type;
        }
        public String getTypeName() {
            return type == TypeClass.OBJECT ? typeStr : type.toString();
        }
    }
}
